// File: evolve.cpp
// Written by Grant Macklem
// 	Grant.Macklem@Colorado.EDU
// 	2:00 PM Recitation, TA: Steve Bjorg
// Last modified on Sunday, November 29, 1998 at 11:52 pm

// This uses the modified winbgim files
/*******************************************************************************
* evolve.exe  A program to simulate evolution of a 9-gene organism.
* Copyright (C) 1998 Grant Macklem
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* You may contact me, the author by emailing Grant.Macklem@Colorado.EDU
*******************************************************************************/


#include "winbgim.h"
#include <iostream.h>			// Provides cin, cout
#include <math.h>					// Provides arctan
#include <time.h>					// Required for randomize
#include <iomanip.h>				// Provides for setw

/* Gene controls are the following
Gene 1: Amount of horizontal movement
Gene 2: Relationship between horizontal and positive vertical (genes 1 and 5)
Gene 3: Length of the next drawn line
Gene 4: Height limiter as recursion deepens
Gene 5: Amount of positive vertical movement
Gene 6: Amount of negative vertical momement
Gene 7: Relationship between horizontal and negative vertical (genes 1 and 6)
Gene 8: Relationship of length of line and negative vertical (genes 3 and 6)
Gene 9: Depth of recursion


Grid is set up like
-------------------------------
|  0 |  1 |  2 |  3 |  4 |  5 |
|-----------------------------|
|  6 |  7 |  8 |  9 | 10 | 11 |
|-----------------------------|
| 12 | 13 | 14 | 15 | 16 | 17 |
|-----------------------------|
          | 18 |
          ------
Where all 18 gene mutations are drawn along with the 19th at the bottom which is
the original, parent gene
*/

// Constants
const double PI = 3.141592653589;

const int WIDTH = 600;
const int HEIGHT = 400;

const int LENGTH = WIDTH / 54;
const int LARGE_LENGTH = WIDTH / 18;

// Directions to draw the biomorph
const int LEFT = 0;
const int RIGHT = 1;

// State variables describing what is currently happening
const int DRAWING = 0;				// program is drawing a picture
const int VIEWING = 1;				// user is viewing gene array
const int FULLSCREEN = 2;			// user has selected to draw a full screen pic
const int VFSCREEN = 3;				// user is viewing the full screen picture
const int GENE_VIEWER = 4;			// user is using the gene viewer program

// Gene Viewer button edges
const int BUTTON[4] = {0.5 * WIDTH / 6 - 30, 3.5 * HEIGHT / 4 - 20,
							  0.5 * WIDTH / 6 + 30, 3.5 * HEIGHT / 4 + 20};

// Structures
struct Point
{
	double x;
	double y;
};

// Prototypes
// Draws each gene by recursively calling itself
void draw_gene
	(Point point,			// Point drawing from
    int direction,		// Direction to draw
    int value[ ],			// Gene values to work with
    double angle,			// Angle which to draw
    double length,		// Length of next line
    int depth);			// Level of recursion

// Function which will add or subtract one from a gene value that the user decided
// to continue evolving.
void left_click (int x, int y);

// Draws the border and lines on the screen as well as calls the functions to
// draw all of the genes
void draw_screen();

// Takes the image the user right clicked on and makes a full screen view
void r_click(int x, int y);

// Randomizes the gene values to create a new gene if the old one looks bad
void new_gene();

// Displays the chromosome information for the full-screened view of the biomorph
void chromosome();

// Draws a rectangle specified by the coordinates given in edges in the desired
// color with the desired label.
void draw_rectangle(const int edges[4], int color, char label1[ ], char label2[ ]);

// Function which will read in gene values and display the resulting biomorph
void gene_viewer();

// Clears the graphic window and displays a message for using the gene viewer
void clear_window();

// Displays instructions to the user about how to run the program
void instructions();

// Instructions to draw a selected biomorph from the muation array fullscreen
void draw_full_screen();



// Global Variables
int gene[9] = {16, 19, 87, 21, 54, 69, 24, 8, 6};
// For interesting ones, try the following.  Each individual gene's influence
// becomes apparent.
// int gene[9] = {82, 86, 67, 76, 4, 37, 11, 11, 9};
int state;			// Current state the program is in (Drawing or viewing)
int fscreen_color;	// Color to draw the full screen picture in
int fscreen_add;		// Amount added to a gene to get the full screen image

void main()
{
   int command;

	randomize();

	register_mouse_handler(WM_LBUTTONDOWN, left_click);
   register_mouse_handler(WM_RBUTTONDOWN, r_click);

	instructions();
   init_window(WIDTH, HEIGHT);

   draw_screen();

	for (;;)		// Infinite loop for program processing
   {
	   while (!kbhit())
		{
			delay(100);
         if (state == DRAWING)
         {
         	cleardevice();
            draw_screen();
            state = VIEWING;
         }
         if (state == FULLSCREEN)
         {
				draw_full_screen();
			}
         if (state == GENE_VIEWER)
         	gene_viewer();
	   }
		command = toupper(getch());
      if ((command == 'R') && (state != VFSCREEN))
      	new_gene();
      if (command == 'Q')
      	exit (EXIT_SUCCESS);
   }	// end for
}	// end main


void instructions()
{
	cout << setw(40 + 7) << "Gene Evolution" << endl;
   cout << setw(56) << "Copyright (C) 1998 Grant Macklem." << endl << endl;
	cout << "   This program will simulate evolution of genes.  It provides a graphical\n"
		  << "representation of the genes as lines.  The aspects of the lines are what the\n"
	     << "genes control.  The program will start off with a sample mutation array.\n"
	     << "This is just a basic biomorph at the bottom with 18 squares above it each\n"
	     << "containing a biomorph mutated from the basic one by changing just one gene\n"
	     << "value by one.  Each row contains six biomorphs.  They are grouped into three\n"
        << "color coded pairs.  The first of each pair has its corresponding gene increased\n"
        << "by one, the second reduced by one.\n\n"
	     << "* Left-click on any picture to set it as the base biomorph and see the\n"
	     << "  mutation array of it.\n"
	     << "* Right-click on any picture to see an enlargement of it.\n"
	     << "* Press 'R' to randomize the genes and create a new base biomorph.\n"
	     << "* Click the 'Gene Viewer' button to be able to enter your own values for\n"
	     << "  each gene and view the result.\n"
	     << "* Press 'Q' on the mutation array screen to quit the program.\n\n"
	     << "Press return to continue." << endl;
	cin.ignore();
}


void draw_full_screen()
{
	Point point;

	cleardevice();
	point.x = WIDTH / 2;
	point.y = 0.64 * HEIGHT;

	// set the color for the gene
	setcolor(fscreen_color);
	draw_gene(point, LEFT, gene, 0, LARGE_LENGTH, 1);
	draw_gene(point, RIGHT, gene, 0, LARGE_LENGTH, 1);
	state = VFSCREEN;
	chromosome();

	// Revert the gene back to the original value
	gene[fscreen_color - 1] -= fscreen_add;
}

void chromosome()
{
	char value[5];
	int text_height, text_width, i;

	setcolor(WHITE);
   text_height = textheight("C");
   outtextxy (20, HEIGHT - text_height, "Chromosome Information: ");
	text_width = textwidth("Chromosome Information: ");

	for (i = 0; i < 9; i++)
   {
		itoa(gene[i], value, 10);
      outtextxy(text_width + 20 * (1 + i), HEIGHT - text_height, value);
   }
}



void draw_screen()
{
	int i, j;				// Row and column loop contrl variables
   int add;					// What number to add (may be negative) to gene value
	int width = WIDTH / 6;			// width of each gene's square
	int height = HEIGHT / 4;			// height of each gene's square
   Point point;

	// Draw gene viewer button
   draw_rectangle(BUTTON, GREEN, "Gene", "Viewer");

	point.x = 2.5 * width;
   point.y = 3.64 * height;

	// Draw borders
   setcolor(BLUE);
	for (i = 0; i <= WIDTH; i+= width)
   	line(i, 0, i, HEIGHT);
   for (i = 0; i <= HEIGHT; i+= height)
   	line(0, i, WIDTH, i);

	// Draw base gene
   setcolor(WHITE);
	draw_gene(point, LEFT, gene, 0, LENGTH, 1);
	draw_gene(point, RIGHT, gene, 0, LENGTH, 1);

   // Draw the rest of the genes
	for(j = 0; j < 3; j++)
   {
   	for(i = 0; i < 6; i++)
      {
			if ((i % 2) == 0)
         	add = 1;
         else
         	add = -1;
         gene[(int) (i / 2) + 3 * j] += add;

			// Set base point to begin drawing the gene
			point.x = (i + 0.5) * width;
         point.y = (j + 0.64) * height;
         setcolor(i / 2 + 3 * j + 1);
			draw_gene(point, LEFT, gene, 0, LENGTH, 1);
			draw_gene(point, RIGHT, gene, 0, LENGTH, 1);
         gene[(int) (i / 2) + 3 * j] -= add;
      }
   }
}

void draw_gene(Point point, int direction, int value[ ], double angle, double length, int depth)

{
	double x, y;				  		// Offsets from the initial point to the new
   										// line end at the beginning of the function and
                                 // the actual new values at the end.
   double new_angle;					// New angle which the line makes with vertical

   x = value[0] + value[1] * value[4] - value[6] * value[5];
	y = value[4] - value[5] - value[3] * depth + value[6] * value[1] - value[7] * value[2];

   if (x == 0)
   	new_angle = PI / 2;
   else
   	new_angle = atan(y / x);

	if (direction == LEFT)
   {
   	new_angle += angle;
   }
   else // direction == RIGHT
   {
   	new_angle = angle - new_angle;
	}

	x = length * sin(new_angle);
	y = length * cos(new_angle);

   line(point.x, point.y, x + point.x, point.y - y);

	point.x += x;
   point.y -= y;

   if (depth < value[8])
   {
		draw_gene(point, LEFT, value, new_angle, length * value[2] / 100, depth + 1);
		draw_gene(point, RIGHT, value, new_angle, length * value[2] / 100, depth + 1);
   }
   else
	{
   	return;
   }

}

void left_click (int x, int y)
{
	int column;								// Column user clicked in
	int row;									// Row user clicked in
   int width = WIDTH / 6;				// width of each gene's square
   int height = HEIGHT / 4;			// height of each gene's square
	int add;									// Amount to add or subtract from gene value

	// User is usine the gene viewer program
   if (state == GENE_VIEWER)
   	return;

	if (state == VFSCREEN)
   {
   	// Return to the mutation array if the user clicks on the screen
      state = DRAWING;
      return;
   }

	// Check to see if the user clicked the gene viewer box.  Since this is outside
   // the gene boxes, the next if statement will return to the previous function
   if ( (x > BUTTON[0]) && (x < BUTTON[2]) && (y > BUTTON[1]) && (y < BUTTON[3]) )
   	state = GENE_VIEWER;

	column = x / width;
	row = y / height;

	// User clicked somewhere else than in the gene boxes on the screen
	if ((column / 2 + 3 * row) > 8)
   	return;

   if ((column % 2) == 0)
		add = 1;
	else
		add = -1;
	gene[column / 2 + 3 * row] += add;
   state = DRAWING;
}

void r_click(int x, int y)
{
	int column;								// Column user clicked in
	int row;									// Row user clicked in
   int width = WIDTH / 6;				// width of each gene's square
   int height = HEIGHT / 4;			// height of each gene's square
	int add;									// Amount to add or subtract from gene value

	// User is usine the gene viewer program
   if (state == GENE_VIEWER)
   	return;

	// User is viewing the full screen picture
	if (state == VFSCREEN)
   	return;

	column = x / width;
	row = y / height;

	// User clicked somewhere else than in the gene boxes on the screen, including
   // the base gene
	if ( (column == 2) && (row == 3) )	// If user clicked base gene
   {
   	state = FULLSCREEN;
      fscreen_color = WHITE;
      fscreen_add = 0;
      return;
   }

	if ((column / 2 + 3 * row) > 8)
   	return;

   if ((column % 2) == 0)
		add = 1;
	else
		add = -1;
	gene[column / 2 + 3 * row] += add;
   state = FULLSCREEN;
   fscreen_color = column / 2 + 3 * row + 1;
   fscreen_add = add;
}

void new_gene()
{
	gene[0] = random(100);
	gene[1] = random(100);
	gene[2] = random(100);
	gene[3] = random(100);
	gene[4] = random(100);
	gene[5] = random(100);
	gene[6] = random(100);
	gene[7] = random(100);
	gene[8] = random(7) + 4;
	state = DRAWING;
}

void draw_rectangle(const int edges[4], int color, char label1[ ], char label2[ ])
{
	int textleft1 = (edges[0] + edges[2] - textwidth(label1)) / 2;
   int texttop1 = (edges[1] + edges[3]) / 2 - textheight(label1);
	int textleft2 = (edges[0] + edges[2] - textwidth(label2)) / 2;
   int texttop2 = (edges[1] + edges[3]) / 2 + 1;

	setfillstyle(SOLID_FILL, color);
   bar(edges[0], edges[1], edges[2], edges[3]);
	outtextxy(textleft1, texttop1, label1);
   outtextxy(textleft2, texttop2, label2);
}

void gene_viewer()
{
	int view[9];		// Temporary gene storage
   int i;				// Loop control variable
   Point point;
   int command;

	for (;;)		// Loop to run gene viewer continuously
	{
   	clear_window();
      clrscr();
		cout << "Enter the values you'd like associated with each gene below.  Any non-integer\n"
	   	  << "value will be converted to an integer before being displayed.  You will have\n"
	        << "a chance to return to the mutation array after the biomorph is drawn." << endl;

		for (i = 0; i < 9; i++)
		{
	   	cout << "Enter gene " << (i + 1) << ": ";
		   cin >> view[i];
	   }

		point.x = WIDTH / 2;
		point.y = 0.64 * HEIGHT;
		//set the color for the gene
		setcolor(WHITE);
      cleardevice();

		draw_gene(point, LEFT, view, 0, LARGE_LENGTH, 1);
		draw_gene(point, RIGHT, view, 0, LARGE_LENGTH, 1);
      outtextxy(20, HEIGHT - 3 * textheight("P '"),
	      "Press 'S' to save the current gene values and return to the mutation array,");
      outtextxy(20, HEIGHT - 2* textheight("P '"),
			"'Q' to return to the mutation array without saving, or any other key to");
      outtextxy(20, HEIGHT - textheight("P '"),
	      "draw another biomorph.");
      command = toupper(getch());
      if (command == 'Q')
		{
			state = DRAWING;
			return;
		}
      if (command == 'S')
      {
      	for (i = 0; i < 9; i++)
         	gene[i] = view[i];
         state = DRAWING;
         return;
      }
   }	// end infinite for loop
}

void clear_window()
{
	cleardevice();
   state = GENE_VIEWER;
   outtextxy(WIDTH / 2 - textwidth("Use text window") / 2,
   		    HEIGHT / 2 - textheight("Use text window") / 2, "Use text window");
}

